/*
  Research carried out within the scope of the Associated International Laboratory: Joint Japanese-French Robotics Laboratory (JRL)
 
  Developed by Mathieu Poirier (LAAS-CNRS)
 
*/


#ifndef HPP_PAT_GEN_H
#define HPP_PAT_GEN_H

/*************************************
INCLUDE
**************************************/

#include <iostream>
#include <fstream>


#include "PatternGeneratorInterface.h"
#include "KineoWorks2/kwsPath.h"
#include "hppDevice.h"

using namespace std ;
using namespace PatternGeneratorJRL;

/*************************************
STRUCTURES
**************************************/

/**
   \brief Structure for a relative footPrint
   \param sx : delta x from the previews footPrint
   \param sy : delta y from the previews footPrint
   \param theta : delta theta from the previews footPrint in degrees
*/
typedef struct TFootPrint
{
  double sx ;
  double sy ;
  double theta ;
}
  TFootPrint ;

/*************************************
CLASS
**************************************/
/**
   \brief hpp Pattern Generator is most an API to use the walkGenJRL.a lib
*/
class ChppWalkPatGen
{

public:

  /**
     \brief  constructor 
     \param  initOpenHRPPosture the Initial OpenHRP Posture for HRP2 using in the Pattern Generator
  */
  ChppWalkPatGen(double initOpenHRPPosture[40]) ;

  /**
     \brief  destructor
  */
  ~ChppWalkPatGen() ;

  /**
     \brief  Initialize a PatternGenerator (interface) with several variables

     omega  : 0.0

     stepheight : 0.07

     singlesupporttime : 0.78

     doublesupporttime : 0.02

     armparameters :  0.5

     LimitsFeasibility : 0.0

     ZMPShiftParameters :  0.015 0.015 0.015 0.015

     TimeDistributionParameters : 2.0 3.7 1.7 3.0

     UpperBodyMotionParameters :  -0.1 -1.0 0.0

     \param aPGI a PatternGeneratorInterface
  */
  void commonInitialization(PatternGeneratorInterface &aPGI);

  /**
     \brief  compute all the Q configuration in 5 ms time for OpenHRP (in FILE .pos, ...) and KINEOWORKS (in structure pathHrp2)
     \param hppDevice : kwsDevice of HRP2
     \param pathHrp2  : kwsPath of HRP2 to be play in the kppInterface
     \param outHRP2Path : vector of HRP2 kwsConfigs equal to HRP2 kwsPath for external use as OpenHrp
     \param outHRP2TimeRef : vector of Time reference for outHRP2Path outHRP2ZMPRef and outHRP2WaistPosition
     \param outHRP2ZMPRef  : vector or ZMP trajectory corresponding to the outHRP2Path
     \param outHRP2WaistPosition : vector or Waist trajectory (position and orientation) corresponding to the outHRP2Path 

     The trajectory of the robot is moved by (ChppWalkPatGen::m_xInit
     ChppWalkPatGen::m_yInit ChppWalkPatGen::m_thetaInit). See
     ChppWalkPatGen::computeAStepSequenceFromAPath for explanation.
  */
  ktStatus computeAllQFromAStepSequence( ChppDeviceShPtr hppDevice, CkwsPathShPtr pathHrp2, 
				     vector<CkwsConfigShPtr> &outHRP2Path, vector<double> &outHRP2TimeRef, 
				     vector<vector<double> > &outHRP2ZMPRef, vector<vector<double> > &outHRP2WaistPosition) ;

  /**
     \brief  compute a stream for Any Curve : step seq : ................. from a step Sequence for PatternGeneratorInterface
     \param aPGI a patternGeneratorInface
  */
  void computeStreamFromStepSequence(PatternGeneratorInterface &aPGI) ;

  /**
     \brief  compute from a box's path all the print foot for HRP2 : this is stored in m_RelativeFootPositions vector
     \param pathBox : the Box's path compute with the HRP2BOX
     \param stepLength : step's length for HRP2 : [0.18 - 0.25] : 0.20 m
     \param stepSpacing : feet's spacing for HRP2 : 0.20 m
     \param lr : used to choose the started foot : left(-1) - right(1)
     \param SECURITYFEET : minimun distance to keep the feet in a safe area  from each other : 0.20 m
  */
  void computeAStepSequenceFromAPath( CkwsPathShPtr pathBox, double stepLength=0.20, double stepSpacing=0.20,  int lr=-1,  double SECURITYFEET = 0.20) ;
 
  /**   
        \brief  compute a step Sequence to do a Half Turn
	\param initConfig : use to set where you want to make the halfTurn (set the varable attXinit, attYinit anf attZinit) 
  */
  void computeAStepSequenceForHalfTurn(CkwsConfigConstShPtr initConfig);


protected :

  /// step sequence of relative foot use to compute the path
  vector<TFootPrint>  relativeFootPrintVector ;
  /// the X shifting use the init config doesn't start from 0
  double attXInit ;
  /// the Y shifting use the init config doesn't start from 0
  double attYInit ;
  /// the Z shifting use the init config doesn't start from 0 (not used yet)
  double attZInit ;
  /// the Theta shifting use the init config doesn't start from 0
  double attThetaInit ;

  /// Initial posture use to configurate Patern Generator (HALFSITTING_POSITION by default)
  double attInitOpenHRPPosturePatGen[40] ;

private :
  ///utility
  double  mod2pi(double a) const ;
  double  angleLimitPi(double angle) const ;
  

};



#endif


