/*
  Research carried out within the scope of the Associated International Laboratory: Joint Japanese-French Robotics Laboratory (JRL)

  Developed by Florent Lamiraux (LAAS-CNRS)

*/




#ifndef HPP_WALK_PLANNER_H
#define HPP_WALK_PLANNER_H

/*************************************
INCLUDE
**************************************/

#include <iostream>

#include "hppPlanner.h"
#include "hppDevice.h"
#include "kwsPlusRoadmap.h"
#include "KineoWorks2/kwsPathOptimizer.h"
#include "KineoWorks2/kwsRoadmapBuilder.h"

#include "KineoWorks2/kwsConfig.h"

using namespace std ;

/*************************************
STRUCTURES
**************************************/
/**
   \brief HALFSITTINGPOSITION  \
 
   legs : 0.0, 0.0, -26.0, 50.0, -24.0, 0.0, 0.0, 0.0, -26.0, 50.0, -24.0, 0.0
 
   chest and head : 0.0, 0.0, 0.0, 0.0, 
 
   right arm : 15.0, -10.0, 0.0, -30.0, 0.0, 0.0, 10.0, 
 
   left arm : 15.0,  10.0, 0.0, -30.0, 0.0, 0.0, 10.0, 
 
   right hand : -10.0, 10.0, -10.0, 10.0, -10.0, 
 
   left hand : -10.0, 10.0, -10.0, 10.0, -10.0  
*/

#define HALFSITTINGPOSITION_DEG_OPENHRP {						\
      0.0, 0.0, -26.0, 50.0, -24.0, 0.0, 0.0, 0.0, -26.0, 50.0, -24.0, 0.0, \
      0.0, 0.0, 0.0, 0.0,						\
      15.0, -10.0, 0.0, -30.0, 0.0, 0.0, 10.0,				\
      15.0,  10.0, 0.0, -30.0, 0.0, 0.0, 10.0,				\
      -10.0, 10.0, -10.0, 10.0, -10.0,					\
      -10.0, 10.0, -10.0, 10.0, -10.0					\
      };

/*************************************
CLASS
**************************************/

#define TOR_OBSTACLE 0.015
#define PENETRATION_BOX 0.01

/**
   \brief Implement a two step walk planning strategy.

   The motion planning strategy is based on an iterative method that plan successive motions. In a first step, a path is computed for the bounding box of the robot and then this path is animated using JRL walk pattern generator. This iterative strategy uses two instanciations of the humanoid robot:
   \li a bounding box moving with a unicycle kinematics and
   \li a fully actuated humanoid robot.
*/
class ChppWalkPlanner : public ChppPlanner {
public: 
  /**
     \brief Constructor
  */
  ChppWalkPlanner() ;

  /**
     \brief Constructor
  */
  ~ChppWalkPlanner() ;

  /**
     \name Problem definition
     @{
  */
  
  /**
     \brief Initialize a Walk Planning problem.
     Compute the bounding box of the first robot in problem  vector (in given configuration) and create a robot with this box.
     
     \param inConfig configuration of the first robot for bounding box computation.

     The box robot is placed in the first problem by a push_front().

  */
  ktStatus initializeProblem(const CkwsConfig& inConfig);
  
  /**
     \brief Set initial configuration of Walk Planning problem
     \param x position and orientation of the robot.
     \param y position and orientation of the robot.
     \param theta position and orientation of the robot.
  */
  ktStatus initConfig(double x, double y, double theta);
  
  /**
     \brief Set goal configuration of Walk Planning problem
     \param x position and orientation of the robot.
     \param y position and orientation of the robot.
     \param theta position and orientation of the robot.
  */
  ktStatus goalConfig(double x, double y, double theta);

  /**
     \brief Set the init OpenHRPposture use in the Pattern Generator in Radian
     \param initPos the Posture (by default HALFSITTING_POSITION)
  */
  ktStatus initOpenHRPPosturePatGenRad(double initPos[40]) ;
  
  /** 
      @}
  */


  /**
     \name Problem resolution
     @{
  */
  /**
     \brief Iteratively solve a walk motion planning problem for a humanoid robot. \     
     \return KD_OK or KD_ERROR
  */
  ktStatus solve();

  /**
     \brief Get the all the element to compute the solution on HRP2. 
     
     \param  outHRP2Path the hpp path of HRP2
     \param  outHRP2TimeRef the Time Reference for all the element
     \param  outHRP2ZMPRef  a vector of ZMPRef(3) ... vector resize on TimeRef size
     \param  outHRP2WaistPosition a vector of WaistPosition(6) ... vector resize on TimeRef size
  
     \return KD_OK or KD_ERROR
  */
  ktStatus solution( vector<CkwsConfigShPtr> &outHRP2Path, vector<double> &outHRP2TimeRef, vector<vector<double> > &outHRP2ZMPRef, vector<vector<double> > &outHRP2WaistPosition );

  /**
     \brief set the bool attDoHalfTurn If attDoHalfTurn = true : the solve function will compute a Half turn for HRP2 
     from the initConfIthProblem and will set the goalConfIthProblem after the half turn.
     Else the solve function will compute a step sequence from the Path of the Boundy Box of HRP2.

     \param b: 
  */
  void doHalfTurn(bool b) ;
  /**
     \brief get the bool attDoHalfTurn
  */
  bool doHalfTurn() ;

  /** 
   *@}
   */


  /**
     \name Search functions
     @{
  */

private:

  /**
     \brief compute the 2st phase : compute all angle for hrp2
     \return KD_OK or KD_ERROR
  */
  ktStatus computePhase2() ;
  /**
     @}
  */


  /**
     \brief Create a robot the body of which is a box.
     The root joint (and only one) is a freeflyer.
     An extra dof is added to the robot in order to use CflicSteeringMethod.
     
  */
  ChppDeviceShPtr createRobotBox(double xMin, double yMin, double zMin,
				 double xMax, double yMax, double zMax); // const;

  /// boolean to set a Half Turn
  bool attDoHalfTurn ;

  /// Initial posture use to configurate Patern Generator (HALFSITTING_POSITION by default)
  double attInitOpenHRPPosturePatGen[40] ;

  /// The compute Path of HRP2
  vector<CkwsConfigShPtr> attHRP2Path ;
  
  /// The time Reference of HRP2 path
  vector<double> attHRP2TimeRef ;
  
  /// The ZMP reference of HRP2 Path
  vector<vector<double> > attHRP2ZMPRef ;
 
  /// the Waist Position of HRP2 Path
  vector<vector<double> > attHRP2WaistPosition ;

};


#endif


