/*
  Research carried out within the scope of the Associated International Laboratory: Joint Japanese-French Robotics Laboratory (JRL)
 
  Developed by Mathieu Poirier (LAAS-CNRS)
 
*/


/*****************************************
 INCLUDES
*******************************************/

#include "hppWalkPatGen.h"
#include "hppOpenHRP/configSwap.h"
#include <sys/time.h>

/*****************************************
 DEFINES
*******************************************/

//utility
#define ABS(x)       (((x) >= 0.0) ? (x) : -(x))
#define FEQ(x,y,f)   (ABS((x)-(y)) < f)


/*****************************************
 METHODS
*******************************************/


// ==========================================================================

ChppWalkPatGen::ChppWalkPatGen(double initOpenHRPPosture[40])
{  
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = initOpenHRPPosture[i];
  }
}

// ==========================================================================

ChppWalkPatGen::~ChppWalkPatGen()
{}

// ==========================================================================

void ChppWalkPatGen::computeAStepSequenceFromAPath( CkwsPathShPtr pathBox, double stepLength,
						    double stepSpacing, int  lr , double SECURITYFEET )
{

  TFootPrint footPrint ;
  double fromStart = 0.0 ;
  double footPlus = 0.0, distanceBetweenFeet = 0.0 ;
  CkwsConfigShPtr aConfig ;
  double xCourbe, yCourbe, thetaCourbe ;
  double xNewAbsolu, yNewAbsolu, thetaNewAbsolu ;
  double xNewRelatif, yNewRelatif, thetaNewRelatif ;
  double xOld, yOld, thetaOld ;
  double x, y, theta ;
  

  //compute the real length of the path
  double lengthPath = pathBox->length() ;
  int nbFullStep = (int) (lengthPath / stepLength) ;

  //-----DEBUG
  //cout << "longueur : " << lengthPath << endl ;
  //cout << "nbFullStep : " << nbFullStep << endl ;
  //cout << "verif : " << stepLength << "*" << nbFullStep << "+" << lastStepLength << "=" << stepLength*nbFullStep<< endl ;
  //------------------------------------------------------------------
 
  // clean the vector
  relativeFootPrintVector.clear() ;

  // first paralelle feet
  xOld = footPrint.sx = 0.0 ;
  yOld = footPrint.sy = lr*stepSpacing/2 ;
  thetaOld = footPrint.theta = 0.0 ;
  relativeFootPrintVector.push_back(footPrint);
  lr = lr * -1 ;

  // initial position of the box. The pattern generator starts from config (0,0,0).
  // We need to move the box path by the inverse of the initial configuration.
  aConfig = pathBox->configAtStart() ;
  attXInit = aConfig->dofValue(1) ;    
  attYInit = aConfig->dofValue(2);
  attZInit = 0 ; // NOT used Yet
  attThetaInit = aConfig->dofValue(6);

  for ( int count=1 ; count <= nbFullStep+1 ; count++) {

    //distance from the start
    fromStart = count * stepLength ;

    if (count == (nbFullStep+1)) {
      //for the last step
      aConfig = pathBox->configAtEnd();
    } else {
      // for the other
      aConfig = pathBox->configAtDistance(fromStart) ;
    }

    x = aConfig->dofValue(1);
    y = aConfig->dofValue(2);
    theta = aConfig->dofValue(6);

    //replace the point in the PatternGenerator Frame
    xCourbe = (x - attXInit)*cos(attThetaInit) + (y - attYInit)*sin(attThetaInit);
    yCourbe = -(x - attXInit)*sin(attThetaInit) + (y - attYInit)*cos(attThetaInit);
    thetaCourbe = theta-attThetaInit;

    // CHECKING - to keep the foot in a security space especialy in curve
    do {
      //compute x, y, theta Absolute of the new foot
      if (lr == 1) {
	// left foot moving
	xNewAbsolu = (stepSpacing+footPlus)/2 * -sin(thetaCourbe) + xCourbe ;
	yNewAbsolu = (stepSpacing+footPlus)/2 * cos(thetaCourbe) + yCourbe ;
      } else {
	//right foot moving
	xNewAbsolu = (stepSpacing+footPlus)/2 *  sin(thetaCourbe) + xCourbe ;
	yNewAbsolu = (stepSpacing+footPlus)/2 *  -cos(thetaCourbe) + yCourbe ;
      }
      distanceBetweenFeet = sqrt(pow(xNewAbsolu-xOld,2) + pow(yNewAbsolu-yOld,2)) ;
      footPlus = footPlus + 0.01 ;
    } while ( distanceBetweenFeet < SECURITYFEET ) ;
    thetaNewAbsolu = thetaCourbe ;

    //compute x, y ,theta Relatif of the new feet from the old feet
    // WARNING : express xNewRelatif , yNewRelatif in the Old foot frame with the orientation of the New foot
    // S=[XNewRelatif, yNewRelatif, thetaNewRelatif] ,  p=[xNewAbsolu, yNewAbsolu], pOld=[xOld, yOld], R(dtheta+thetaOld)= rotation matrix
    // R * S = (p - pOld)
    // BECAUSE : WPG first compute the rotation of the foot then the translation.
    xNewRelatif = cos(thetaNewAbsolu)*(xNewAbsolu-xOld) + sin(thetaNewAbsolu)*(yNewAbsolu-yOld) ;
    yNewRelatif = -sin(thetaNewAbsolu)*(xNewAbsolu-xOld) + cos(thetaNewAbsolu)*(yNewAbsolu-yOld) ;
    thetaNewRelatif = thetaNewAbsolu - thetaOld ;
    thetaNewRelatif = mod2pi(thetaNewRelatif) ;   // between [0;2Pi] 
    thetaNewRelatif = angleLimitPi(thetaNewRelatif) ; // between [-Pi;Pi]
    
    //stock the info in footPrint
    footPrint.sx = xNewRelatif ;
    footPrint.sy = yNewRelatif ;
    footPrint.theta = thetaNewRelatif*180/M_PI ;  // en degree
    relativeFootPrintVector.push_back(footPrint);
    lr = lr * -1 ;

    //stock the info in Old
    xOld = xNewAbsolu ;
    yOld = yNewAbsolu ;
    thetaOld = thetaNewAbsolu ;
    footPlus = 0.0 ;

  }

  //finally put feet forward parallele
  footPrint.sx = 0.0 ;
  footPrint.sy = lr*stepSpacing ;
  footPrint.theta = 0.0 ;  /* en degree */
  relativeFootPrintVector.push_back(footPrint);
  lr = lr * -1 ;

}

// ==========================================================================

void ChppWalkPatGen::commonInitialization(PatternGeneratorInterface &aPGI)
{

 const char lBuffer[9][256] =
    {":omega 0.0",
     ":stepheight 0.07",
     ":singlesupporttime 0.78",
     ":doublesupporttime 0.02",
     ":armparameters 0.5",
     ":LimitsFeasibility 0.0",
     ":ZMPShiftParameters 0.015 0.015 0.015 0.015",
     ":TimeDistributionParameters 2.0 3.7 1.7 3.0",
     ":UpperBodyMotionParameters -0.1 -1.0 0.0"
    };
  
  for(int i=0;i<9;i++)
    {
      std::istringstream strm(lBuffer[i]);
      aPGI.ParseCmd(strm);
    }

}

// ==========================================================================

void ChppWalkPatGen::computeAStepSequenceForHalfTurn(CkwsConfigConstShPtr initConfig) {

  // X
  attXInit = initConfig->dofValue(0) ;
  // Y
  attYInit = initConfig->dofValue(1) ;
  // Z 
  attZInit = initConfig->dofValue(2) ;
  
  istringstream strm; 

  // HALF TURN in 6 step and 30 degree
  strm.str(" 0.0 -0.10 0.0 \
             -0.0848 0.2048   -30.0			  \
             0.0848  -0.2048  -30.0			  \
             -0.0848 0.2048   -30.0			  \
             0.0848  -0.2048  -30.0			  \
             -0.0848 0.2048   -30.0			  \
             0.0848  -0.2048  -30.0			  \
             0.0 0.20 0.0");

  // clean the vector
  relativeFootPrintVector.clear() ;

  // build the stepSequence
  double sx, sy, theta ;
  TFootPrint footPrint ;

  for (unsigned int count= 0 ; count < 8 ; count ++) {
    strm >> sx ;
    strm >> sy ;
    strm >> theta ;
    footPrint.sx = sx ;
    footPrint.sy = sy ;
    footPrint.theta = theta ; 
    relativeFootPrintVector.push_back(footPrint);
  }
}

// ==========================================================================

void ChppWalkPatGen::computeStreamFromStepSequence(PatternGeneratorInterface &aPGI)
{
  commonInitialization(aPGI) ;

  istringstream strm2 ;
  stringstream strm ;

  strm << ":stepseq " ;
  for (unsigned int count =0 ;count < relativeFootPrintVector.size() ; count++) {
    strm <<"\t" << relativeFootPrintVector[count].sx <<"\t" << relativeFootPrintVector[count].sy <<"\t" << relativeFootPrintVector[count].theta ;
  }

  strm2.str(strm.str()) ;

  aPGI.ParseCmd(strm2);
}

// ==========================================================================

ktStatus ChppWalkPatGen::computeAllQFromAStepSequence( ChppDeviceShPtr hppDevice, CkwsPathShPtr kwsPathHRP2,
						   vector<CkwsConfigShPtr> &HRP2Path, vector<double> &HRP2TimeRef, 
						   vector<vector<double> > &HRP2ZMPRef, vector<vector<double> > &HRP2WaistPosition)
{
  // collision status
  ktStatus col_state = KD_OK;

  // make sure that the data correspond to the last path computed
  HRP2WaistPosition.clear() ;
  HRP2ZMPRef.clear() ;
  HRP2TimeRef.clear() ;
  HRP2Path.clear() ;

  CkwsConfigShPtr startConfig=CkwsConfig::create(hppDevice) ;
  CkwsSMLinearShPtr steeringMethod = CkwsSMLinear::create();
  bool firstTest = true ;

  // IMPORTANTE NOTE :
  // This program takes 5 arguments :
  // PATH_TO_PREVIEWCCONTRAL_PARAMS_FILE 
  // PATH_TO_VRML_FILE 
  // VRML_FILE_NAME 
  // PATH_TO_SPECIFICITIES_XML 
  // LINK_JOINT_RANK" << endl;

  string Global= WALKGENJRL_DATA "/PreviewControlParameters.ini";
  Global+= " ";
  Global+= OPENHRP_PREFIX;
  Global+="/etc/HRP2JRL/" ;
  Global+= " ";
  Global+="HRP2JRLmain.wrl" ;
  Global+= " ";
  Global+= WALKGENJRL_DATA "/HRP2Specificities.xml";
  Global+= " ";
  Global+= WALKGENJRL_DATA "/HRP2LinkJointRank.xml";
  
  std::istringstream strm(Global);
 
  PatternGeneratorInterface * aPGI;
  aPGI = new PatternGeneratorInterface(strm);

  //Set initial Position - half sitting position
  MAL_VECTOR_DIM(initialPosition,double,40);
  for(unsigned int i=0;i<MAL_VECTOR_SIZE(initialPosition);i++) 
    initialPosition(i) =  attInitOpenHRPPosturePatGen[i];
  aPGI->SetCurrentJointValues(initialPosition);

  // Set the walking mode:
  //     - 0 : Normal mode.
  //     - 1 : Waist height variation.
  //     - 3 : Kineo works mode.
  //     - 4 : Following the upper-body motion provided by another plugin.
  // Specify the walking mode: here the default one.
  istringstream strm2(":walkmode 0");
  aPGI->ParseCmd(strm2);
  

  computeStreamFromStepSequence(*aPGI) ;


  //debug - computation time
  struct timeval begin,end;
  gettimeofday(&begin,0);

  unsigned long int NbOfIt=0;
  double time=0.0 ;

  MAL_VECTOR_DIM(currentVelocity,double,46);
  MAL_VECTOR_DIM(currentConfiguration,double,46);
  MAL_VECTOR_DIM(ZMPTarget,double,3);
 


  // Main Loop of PatternGenerator Computation
  while(aPGI->RunOneStepOfTheControlLoop(currentConfiguration,
					 currentVelocity,
					 ZMPTarget)) {

    NbOfIt++;

   

    // time for file : every 0.005 secondes
    time = NbOfIt*0.005 ;

    // shift all the freeflyer position if HRP2 did not start from 0, 0
    double xPosShifted, yPosShifted, zPosShifted, rollPosShifted, pitchPosShifted, yawPosShifted ;
    xPosShifted = currentConfiguration(0)*cos(attThetaInit) - currentConfiguration(1)*sin(attThetaInit) + attXInit ;
    yPosShifted = currentConfiguration(0)*sin(attThetaInit) + currentConfiguration(1)*cos(attThetaInit) + attYInit ;
    zPosShifted = currentConfiguration(2)+ attZInit ; 
    rollPosShifted = currentConfiguration(3) ; 
    pitchPosShifted = currentConfiguration(4) ; 
    yawPosShifted = currentConfiguration(5) + attThetaInit;

    // ------------------------ Write in Data Structures ---------------//
   
    vector<double> currentZMPTarget(3) ;
    vector<double> currentWaistPosition(6) ;
    vector<double> currentOpenHrpConfig(40) ;

    HRP2TimeRef .push_back(time) ;

    currentZMPTarget[0] = ZMPTarget(0) ; currentZMPTarget[1] = ZMPTarget(1) ; currentZMPTarget[2] = ZMPTarget(2) ;
    HRP2ZMPRef.push_back(currentZMPTarget) ;

    currentWaistPosition[0] =  xPosShifted ; currentWaistPosition[1] = yPosShifted ; currentWaistPosition[2] = zPosShifted ; 
    currentWaistPosition[3] =  rollPosShifted ; currentWaistPosition[4] = pitchPosShifted ; currentWaistPosition[5] = yawPosShifted ;
    HRP2WaistPosition.push_back(currentWaistPosition) ;

   // Becareful do not insert currentConfiguration(0)->currentConfiguration(5) : because they are position and orientation 
    for( unsigned int count = 6 ; count <  MAL_VECTOR_SIZE(currentConfiguration) ; count++) {
      currentOpenHrpConfig[count-6] = currentConfiguration(count) ;
    }

    //swap OpenHRP config to Hpp config -- see package hppOpenHRP configswap.h
    CkwsConfigShPtr endConfig  =CkwsConfig::create(hppDevice) ;
    toHppConfig (currentOpenHrpConfig, currentWaistPosition , endConfig ) ;

    HRP2Path.push_back(endConfig) ;   		


    // -----------------------------------------------------------------------------//
   
    if(firstTest) {

      for (unsigned int z=0 ; z < endConfig->size() ; z++ ) {
	startConfig->dofValue(z, endConfig->dofValue(z)) ;
      }

      firstTest = false ;
    } else {

      // note : Kws can't built a Direct Path with startConfig = endConfig
      //creation of the direct path
      CkwsDPLinearShPtr DPLinear = CkwsDPLinear::create(*startConfig, *endConfig, steeringMethod) ;

      // check if the path is collising once for 2 configuration
      if(NbOfIt % 2 == 0){
	hppDevice->setCurrentConfig(*endConfig);
	if(hppDevice->collisionTest()){
	  cout<<"(self-)collision detected at t="<<NbOfIt*0.005<<endl;
	  kwsPathHRP2->clear();
	  return KD_ERROR;
	}
      }
      
      //add the direct path to the path
      kwsPathHRP2->appendDirectPath(DPLinear) ;

      for (unsigned int z=0 ; z < endConfig->size() ; z++ ) {
	startConfig->dofValue(z, endConfig->dofValue(z)) ;
      }
    }
  }

 
  //debug computation of time
  gettimeofday(&end,0);
  double ltime = end.tv_sec-begin.tv_sec + 0.000001 * (end.tv_usec - begin.tv_usec);  
  cout << endl << "Number of iterations " << NbOfIt << endl;
  cout << "Time consumption: " << (double)ltime/(double)NbOfIt << endl;

  return col_state;
}


// ==========================================================================



double  ChppWalkPatGen::mod2pi(double a) const
{
  return((double)((a=fmod(a, 2*M_PI))>0.0)?a:(FEQ(a,0.0,1e-10)?0.0:(a + 2*M_PI)));
}


// ==========================================================================

double  ChppWalkPatGen::angleLimitPi(double angle) const
{
  while (angle < -M_PI){
    angle += 2*M_PI;
  }
  while (angle > M_PI){
    angle -= 2*M_PI;
  }
  return angle;

}
