/*
  Research carried out within the scope of the Associated International Laboratory: Joint Japanese-French Robotics Laboratory (JRL)
 
  Developed by Florent Lamiraux (LAAS-CNRS)
  and Mathieu Poirier (LAAS-CNRS)
*/

/*! \addtogroup walk
 *@{
 */


/*****************************************
 INCLUDES
*******************************************/

#include <iostream>
#include <fstream>
#include "hppWalkPlanner.h"
#include "hppWalkPatGen.h"
#include "hppBody.h"
#include "flicSteeringMethod.h"



#include "KineoModel/kppRotationJointComponent.h"
#include "KineoModel/kppFreeFlyerJointComponent.h"
#include "KineoModel/kppAnchorJointComponent.h"
#include "KineoModel/kppJointComponent.h"
#include "KineoModel/kppSolidComponentRef.h"

#include "hppBox.h"

#define OPT_TIMES   4
#define OPT_MAX_NB  3
// ==========================================================================


ChppWalkPlanner::ChppWalkPlanner():ChppPlanner(){

      
  attHRP2Path.clear() ;
  attHRP2TimeRef.clear() ;
  attHRP2ZMPRef.clear() ;
  attHRP2WaistPosition.clear() ;

  
  attDoHalfTurn = false ;

  double temp[40] =  HALFSITTINGPOSITION_DEG_OPENHRP ;  
  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = temp[i]*M_PI/180.0;
  }

}

// ==========================================================================


ChppWalkPlanner::~ChppWalkPlanner(){

 
  attHRP2Path.clear() ;
  attHRP2TimeRef.clear() ;
  attHRP2ZMPRef.clear() ;
  attHRP2WaistPosition.clear() ;

}

// ==========================================================================

ktStatus ChppWalkPlanner::initializeProblem(const CkwsConfig& inConfig)
{

  double xMin=0, yMin=0, zMin=0, xMax=0, yMax=0, zMax=0;
  unsigned int nbRobots = getNbHppProblems();

  

  if (nbRobots == 0) {
    cerr << "ERROR - ChppWalkPlanner::initializeProblem: there is no robot." << endl;
    return KD_ERROR;
  }

  // Assume humanoid robot is at rank 0 in problem vector.
  ChppDeviceShPtr humanoidRobot = KIT_DYNAMIC_PTR_CAST(ChppDevice, hppProblemVector[0].getRobot());
  humanoidRobot->setCurrentConfig(inConfig);
   
  // Compute bouding box of humanoid robot.
  if (humanoidRobot->axisAlignedBoundingBox(xMin, yMin, zMin, xMax, yMax, zMax) != KD_OK) {
    cerr << "ERROR - ChppWalkPlanner::initializeProblem: robot seems to have no geometry" << endl;
    return KD_ERROR;
  }
  const ChppDeviceShPtr& robotBox = createRobotBox(xMin, yMin, zMin, xMax, yMax, zMax);
  if (!robotBox) {
    cerr << "ERROR - ChppWalkPlanner::initializeProblem: failed to create robot bounding box." << endl;    
    return KD_ERROR;
  }

  //make the joint invisible by default
  robotBox->isVisible(false) ;
 
  //the body of the bouding box transparent
  robotBox->solidComponentRef(0)->referencedSolidComponent()->isTransparent(true) ;

  // Register bounding box in problem vector.
  addHppProblemAtBeginning(robotBox);

  // Set default tools for the robot bouding box.
  // Create steering method
  CflicSteeringMethodShPtr steeringMethod = CflicSteeringMethod::create() ; 
  if (!steeringMethod)
    return KD_ERROR;
  steeringMethodIthProblem(0, steeringMethod);

  //create RoadMap
  CkwsRoadmapShPtr kwsPlusRoadmap = CkwsRoadmap::create(robotBox) ;
  if (!kwsPlusRoadmap)
    return KD_ERROR;

  // Create a roadmap builder strategie
  CkwsBasicRdmBuilderShPtr kwsRoadmapBuilder = CkwsBasicRdmBuilder::create(kwsPlusRoadmap, PENETRATION_BOX);
  //CkwsDiffusingRdmBuilderShPtr kwsRoadmapBuilder = CkwsDiffusingRdmBuilder::create(kwsPlusRoadmap, PENETRATION_BOX);
  //CkwsIPPRdmBuilderShPtr kwsRoadmapBuilder = CkwsIPPRdmBuilder::create(kwsPlusRoadmap, PENETRATION_BOX) ;

  if (!kwsRoadmapBuilder)
    return KD_ERROR;
  roadmapBuilderIthProblem(0, kwsRoadmapBuilder);

  // Create a Path Optimizer
  // CkwsRandomOptimizerShPtr kwsOptimizer = CkwsRandomOptimizer::create(robotBox->specificDistance());
  // kwsOptimizer->doesClear(true);  

  CkwsAdaptiveShortcutOptimizerShPtr kwsOptimizer =  CkwsAdaptiveShortcutOptimizer::create(robotBox->specificDistance());

  kwsOptimizer->shortcutMethod(CkwsShortcutDirect::create());

  kwsOptimizer->maxNbLoop(OPT_MAX_NB);
  cout<<" max nb loop for the optimizer "<< kwsOptimizer->maxNbLoop ()<<endl;

  pathOptimizerIthProblem(0, kwsOptimizer) ;

  std::vector<CkcdObjectShPtr> oList = obstacleList();
  for(unsigned int i =0; i<oList.size(); i++){
    oList[i]->tolerance(TOR_OBSTACLE);
    cout<<"tolerance "<<TOR_OBSTACLE<<" set to obstacle "<<i<<endl;
  }

  cout << endl << endl << " DEBUG 3 " << endl << endl ;
  return KD_OK;

}

// ==========================================================================

ktStatus ChppWalkPlanner::initConfig(double inX, double inY, double inTheta)
{

  if (getNbHppProblems() < 2) {
    std::cerr << "ERROR - ChppWalkPlanner::initConfig: not enought robot has been defined : " <<getNbHppProblems() << std::endl;
    return KD_ERROR;
  }

  /*
    Build configuration for bounding box.
  */

  const ChppDeviceShPtr boundingBox = KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(0));
 
  if( !boundingBox){
    std::cerr << "ERROR - ChppWalkPlanner::initConfig: the bounding box is empty" << std::endl ;
  }
  
  std::vector<double> dofValue;
  dofValue.push_back(0.0);
  dofValue.push_back(inX);
  dofValue.push_back(inY);
  dofValue.push_back(0.0);
  dofValue.push_back(0.0);
  dofValue.push_back(0.0);
  dofValue.push_back(inTheta);
  
  CkwsConfigShPtr initConfBox =  CkwsConfig::create(boundingBox, dofValue);

  if (!initConfBox) {
    std::cerr << "ERROR - ChppWalkPlanner::initConfig: could not createthe init  configuration boundingBox." << std::endl;
    return KD_ERROR;
  }
    
  if (initConfIthProblem(0, initConfBox) == KD_ERROR) {
    std::cerr << "ERROR - ChppWalkPlanner::initConfig could not set initial configuration of robot box." << std::endl;
    return KD_ERROR;
  }

  /*
    Build configuration for HRP2.
  */

  const ChppDeviceShPtr HRP2Device = KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(1));

  if( !HRP2Device){
    std::cerr << "ERROR - ChppWalkPlanner::initConfig: HRP2 is empty" << std::endl ;
  }

  CkwsConfigShPtr  initConfHRP2 =  CkwsConfig::create(HRP2Device);
  initConfHRP2->dofValue(0, inX);
  initConfHRP2->dofValue(1, inY);
  initConfHRP2->dofValue(2, 0);
  initConfHRP2->dofValue(3, 0);
  initConfHRP2->dofValue(4, 0);
  initConfHRP2->dofValue(5, inTheta);
  
  if (!initConfHRP2) {
    std::cerr << "ERROR - ChppWalkPlanner::initConfig: could not create the init configuration HRP2." << std::endl;
    return KD_ERROR;
  }
    
  if (initConfIthProblem(1, initConfHRP2) == KD_ERROR) {
    std::cerr << "ERROR - ChppWalkPlanner::initConfig could not set initial configuration of robot HRP2." << std::endl;
    return KD_ERROR;
  }

  return KD_OK;
}

// ==========================================================================

ktStatus ChppWalkPlanner::goalConfig(double inX, double inY, double inTheta)
{
  if (getNbHppProblems() < 2) {
    std::cerr << "ERROR - ChppWalkPlanner::goalConfig - not enought robot has been defined : " << getNbHppProblems() << std::endl;
    return KD_ERROR;
  }

  /*
    Build configuration for bounding box.
  */  

  const ChppDeviceShPtr boundingBox = KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(0));
  
  if( !boundingBox){
    std::cerr << "ERROR - ChppWalkPlanner::goalConfig: the bounding box is empty" << std::endl ;
  }

  std::vector<double> dofValue;
  dofValue.push_back(0.0);
  dofValue.push_back(inX);
  dofValue.push_back(inY);
  dofValue.push_back(0.0);
  dofValue.push_back(0.0);
  dofValue.push_back(0.0);
  dofValue.push_back(inTheta);
  
  CkwsConfigShPtr goalConfBox =  CkwsConfig::create(boundingBox, dofValue);
  
  if (!goalConfBox) {
    std::cerr << "ERROR - ChppWalkPlanner::goalConfig : could not create the goal  configuration boundingBox." << std::endl;
    return KD_ERROR;
  }
    
  if (goalConfIthProblem(0, goalConfBox) == KD_ERROR) {
    std::cerr << "ERROR - ChppWalkPlanner::goalConfig :could not set initial configuration of robot box." << std::endl;
    return KD_ERROR;
  }
 
  /*
    Build configuration for HRP2.
  */

  const ChppDeviceShPtr HRP2Device = KIT_DYNAMIC_PTR_CAST(ChppDevice, robotIthProblem(1));
  
  if( !HRP2Device){
    std::cerr << "ERROR - ChppWalkPlanner::goalConfig: HRP2 is empty" << std::endl ;
  }

  CkwsConfigShPtr  goalConfHRP2 =  CkwsConfig::create(HRP2Device);
  goalConfHRP2->dofValue(0, inX);
  goalConfHRP2->dofValue(1, inY);
  goalConfHRP2->dofValue(2, 0);
  goalConfHRP2->dofValue(3, 0);
  goalConfHRP2->dofValue(4, 0);
  goalConfHRP2->dofValue(5, inTheta);
    
  if (!goalConfHRP2) {
    std::cerr << "ERROR - ChppWalkPlanner::goalConfig: could not create the goal configuration HRP2." << std::endl;
    return KD_ERROR;
  }
    
  if (goalConfIthProblem(1, goalConfHRP2) == KD_ERROR) {
    std::cerr << "ERROR - ChppWalkPlanner::goalConfig could not set initial configuration of robot HRP2." << std::endl;
    return KD_ERROR;
  }

  return KD_OK;
}

// ==========================================================================


ktStatus ChppWalkPlanner::initOpenHRPPosturePatGenRad(double initPos[40]) {

  for(unsigned int i = 0 ; i < 40 ; i++) {
    attInitOpenHRPPosturePatGen[i] = initPos[i];
  }
  return KD_OK;
}

// ==========================================================================


ktStatus ChppWalkPlanner::solve()
{
  ktStatus status ;
  
  
  cout << "_______________________ PHASE 1 ____________________________" << endl ;
  
 
  if (!robotIthProblem(0)) {
    cerr << " ERROR - ChppWalkPlanner::solve() HRP2 BOX not Loaded" << endl ;
    return KD_ERROR ;
  }

  if(!initConfIthProblem(0)){
    cerr << "ERROR -  ChppWalkPlanner::solve() Init configuration not Initialized" << endl ; 
    return KD_ERROR ;
  }

  if(!goalConfIthProblem(0)){
    cerr << "ERROR -  ChppWalkPlanner::solve() goal configuration not Initialized" << endl ; 
    return KD_ERROR ;
  }


  // desactived HRP2 and box collision 
  CkppDeviceComponentShPtr kppRobot0 = robotIthProblem(0);
  CkppDeviceComponentShPtr kppRobot1 = robotIthProblem(1);
  ChppDeviceShPtr robot0 = KIT_DYNAMIC_PTR_CAST(ChppDevice, kppRobot0);
  ChppDeviceShPtr robot1 = KIT_DYNAMIC_PTR_CAST(ChppDevice, kppRobot1);
  robot0->ignoreDeviceForCollision( robot1 ) ;
  robot1->ignoreDeviceForCollision( robot0 ) ;
  
  //
  //  TO TEST Only A DirectPath
  //
  //   CkwsPathShPtr myPath(CkwsPath::create(robot0)) ;
  //   CkwsDirectPathShPtr myDP ;
  
  //   myDP = steeringMethodIthProblem(0)->makeDirectPath( *(initConfIthProblem(0)), *(goalConfIthProblem(0)) ) ;
  
  //   myPath->appendDirectPath(myDP) ;
  //   hppProblemVector[0].addPath(myPath) ;

  //
  // For Roadmap
  // 
  if (attDoHalfTurn == false){
    status = solveOneProblem(0);
    if (status == KD_ERROR) {
      return status;
    }
  }

  // optimize again OPT_TIMES times if not half turn.
  if (attDoHalfTurn == false){
    if (hppProblemVector[0].pathOptimizer()) {

      // get the optimized path in solveOneProblem()
      CkwsPathShPtr kwsPath = getPath(0 , getNbPaths(0)-1 );

      cout << "ChppPlanner::solveOneProblem: optimizing with roadmap penetration "
	   <<hppProblemVector[0].roadmapBuilder()->penetration()<<" ... " << endl;

      for(int i=0; i<OPT_TIMES; i++){
	cout<<i<<" ";

	hppProblemVector[0].pathOptimizer()->optimizePath(kwsPath, hppProblemVector[0].roadmapBuilder()->penetration());
	cout<<" number of direct path: "<<kwsPath->countDirectPaths()<<endl;

      }
      cout<<" ... path optimization finished. "<<endl;

      hppProblemVector[0].addPath(kwsPath);

    } else {
      cerr << " no Optimizer Defined " << endl ;
    }
  }

  cout << "_______________________ PHASE 2 ____________________________" << endl ;
  

  status = computePhase2() ;
  
  cout << "_______________________ PHASE 3 ____________________________" << endl ;
  
  
  return status ;
}


// ==========================================================================

ktStatus ChppWalkPlanner::computePhase2( )
{
  ktStatus status = KD_OK;

  //some checking
  if(hppProblemVector.size() < 2) {
    cout << " ERROR - ChppWalkPlanner::computePhase2() : NO Device in vector[1] " << endl ;
    return KD_ERROR ;
  }

  if(hppProblemVector[1].getRobot()->name() == "HRP2") {
    cout << " ERROR - ChppWalkPlanner::computePhase2() : the Loaded device is not HRP2 " << endl ;
    return KD_ERROR ;
  }


  ChppWalkPatGen patGen(attInitOpenHRPPosturePatGen) ;

  CkwsPathShPtr pathHrp2 = CkwsPath::create(hppProblemVector[1].getRobot()) ;
  CkwsPathShPtr pathBox = getPath(0 , getNbPaths(0)-1 );

  if (attDoHalfTurn == false) {
    // compute from a path
    patGen.computeAStepSequenceFromAPath(pathBox) ;
  }
  else {
    // make a half turn
    patGen.computeAStepSequenceForHalfTurn(initConfIthProblem(1)) ;
  }
  if(patGen.computeAllQFromAStepSequence(KIT_DYNAMIC_PTR_CAST(ChppDevice, hppProblemVector[1].getRobot()),  
					 pathHrp2, attHRP2Path, 
					 attHRP2TimeRef, attHRP2ZMPRef, attHRP2WaistPosition)){
    cout<<" Detected collision in walking motion "<<endl;
    return KD_ERROR;
  }

  // put the REAL Config in init and goal configs of HRP2 
  initConfIthProblem(1,pathHrp2->configAtStart()); 
  goalConfIthProblem(1,pathHrp2->configAtEnd()) ;
 
  cout << "nbre de DP dans path : " << pathHrp2->countDirectPaths() <<endl;
  hppProblemVector[1].addPath(pathHrp2) ;   

  return KD_OK;

}

// ==========================================================================

ktStatus ChppWalkPlanner::solution( vector<CkwsConfigShPtr> &outHRP2Path, vector<double> &outHRP2TimeRef, 
		   vector<vector<double> > &outHRP2ZMPRef, vector<vector<double> > &outHRP2WaistPosition ) {

  if(attHRP2Path.empty()) {
    cerr << " ERROR : ChppWalkPlanner::solution - attHRP2Path empty " << endl ;
    return KD_ERROR ;
  }
  else outHRP2Path = attHRP2Path ;

  if(attHRP2TimeRef.empty()) {
    cerr << " ERROR : ChppWalkPlanner::solution - attHRP2TimeRef empty " << endl ;
    return KD_ERROR ;
  }
  else outHRP2TimeRef = attHRP2TimeRef ;

  if(attHRP2ZMPRef.empty()) {
    cerr << " ERROR : ChppWalkPlanner::solution - attHRP2ZMPRef empty " << endl ;
    return KD_ERROR ;
  }
  else outHRP2ZMPRef = attHRP2ZMPRef  ;

  if(attHRP2WaistPosition.empty()) {
    cerr << " ERROR : ChppWalkPlanner::solution - attHRP2WaistPosition empty " << endl ;
    return KD_ERROR ;
  }
  else outHRP2WaistPosition = attHRP2WaistPosition ;

   return KD_OK ;

}

// ==========================================================================

void  ChppWalkPlanner::doHalfTurn(bool b){
  attDoHalfTurn = b ;
}

// ==========================================================================

bool  ChppWalkPlanner::doHalfTurn(){
  return attDoHalfTurn ;
}


// ==========================================================================

ChppDeviceShPtr ChppWalkPlanner::createRobotBox(double xMin, double yMin, double zMin,
						double xMax, double yMax, double zMax) // const
{

  ChppDeviceShPtr hppDevice = ChppDevice::create("Humanoid Box");

  CkitMat4 absPos;
  ChppBoxShPtr polyBox = ChppBox::create("H.Box",xMax-xMin, yMax-yMin, zMax-zMin);
  polyBox->diffuseColor( CkppColor( 0.2f, 0.2f, 0.6f, 0.4f ) );
  absPos.translate(.5*(xMin+xMax), .5*(yMin+yMax), .5*(zMin+zMax));
  polyBox->setAbsolutePosition(absPos);
  polyBox->makeCollisionItem();
  
  
  // create ExtraDof
  CkwsDofShPtr kwsExtraDof ;
  kwsExtraDof = CkwsDof::create(false);
  // Bound curvature
  //kwsExtraDof->bounds(-1.0, 1.0);
  //kwsExtraDof->isBounded(true);
  kwsExtraDof->isLocked(true) ;
  hppDevice->addExtraDof(kwsExtraDof);
  
  
  // build plan root joint
  CkppJointComponentShPtr kppJoint = CkppFreeFlyerJointComponent::create( "Rootjoint.H.Box" );
  // Bound joint
  // X
  kppJoint->kwsJoint()->dof(0)->isBounded(true);
  kppJoint->kwsJoint()->dof(0)->bounds(-0.5,2.5); 
  // Y
  kppJoint->kwsJoint()->dof(1)->isBounded(true);
  kppJoint->kwsJoint()->dof(1)->bounds(-1.5,1.5); 
  // Z
  kppJoint->kwsJoint()->dof(2)->isLocked(true);
  // Rx
  kppJoint->kwsJoint()->dof(3)->isLocked(true);
  // Ry
  kppJoint->kwsJoint()->dof(4)->isLocked(true);
  // Rz
  kppJoint->kwsJoint()->dof(5)->isBounded(true);
  kppJoint->kwsJoint()->dof(5)->bounds(-2*M_PI,2*M_PI);

  //attach the body to the joint 

  // Create and attach hppBody to joint
  ChppBodyShPtr hppBody = ChppBody::create(polyBox->name());

  kppJoint->kwsJoint()->setAttachedBody( (CkwsBodyShPtr) hppBody);

  kppJoint->addSolidComponentRef( CkppSolidComponentRef::create(polyBox) ); 
  kppJoint->doesDisplayPath( true );

  //attach the joint to the device
  hppDevice->rootJointComponent(kppJoint);

  // add obstacles to polyBox
  // inner object with absPos  and outer object from obstacleList.

  // create collision list of body
  std::vector<CkcdObjectShPtr> collisionList ;
  // inner objects matrix list
  std::vector< CkitMat4 > matList;
  collisionList.clear();
  matList.clear();

  collisionList.push_back(polyBox);
  matList.push_back(absPos);

  hppBody->setInnerObjects(collisionList);
  hppBody->innerObjects(collisionList, matList);

  hppBody->setOuterObjects(obstacleList());

  return hppDevice ;
  
}


